@@ -1,6 +1,7 @@
 #include "pgroonga.h"
 
 #include "pgrn-column-name.h"
+#include "pgrn-compatible.h"
 #include "pgrn-convert.h"
 #include "pgrn-global.h"
 #include "pgrn-groonga.h"
@@ -214,11 +215,12 @@ PGrnLookupWithSize(const char *name,
 {
 	grn_obj *object;
 	object = grn_ctx_get(ctx, name, nameSize);
-	if (!object)
+	if (!object && errorLevel != PGRN_ERROR_LEVEL_IGNORE)
 	{
-		PGrnCheckRC(GRN_INVALID_ARGUMENT,
-					"object isn't found: <%.*s>",
-					(int)nameSize, name);
+		PGrnCheckRCLevel(GRN_INVALID_ARGUMENT,
+						 errorLevel,
+						 "object isn't found: <%.*s>",
+						 (int)nameSize, name);
 	}
 	return object;
 }
@@ -264,7 +266,7 @@ PGrnLookupSourcesTable(Relation index, int errorLevel)
 
 	snprintf(name, sizeof(name),
 			 PGrnSourcesTableNameFormat,
-			 index->rd_node.relNode);
+			 PGRN_RELATION_GET_LOCATOR_NUMBER(index));
 	return PGrnLookup(name, errorLevel);
 }
 
@@ -275,7 +277,7 @@ PGrnLookupSourcesCtidColumn(Relation index, int errorLevel)
 
 	snprintf(name, sizeof(name),
 			 PGrnSourcesTableNameFormat "." PGrnSourcesCtidColumnName,
-			 index->rd_node.relNode);
+			 PGRN_RELATION_GET_LOCATOR_NUMBER(index));
 	return PGrnLookup(name, errorLevel);
 }
 
@@ -286,7 +288,7 @@ PGrnLookupLexicon(Relation index, unsigned int nthAttribute, int errorLevel)
 
 	snprintf(name, sizeof(name),
 			 PGrnLexiconNameFormat,
-			 index->rd_node.relNode,
+			 PGRN_RELATION_GET_LOCATOR_NUMBER(index),
 			 nthAttribute);
 	return PGrnLookup(name, errorLevel);
 }
@@ -298,7 +300,7 @@ PGrnLookupIndexColumn(Relation index, unsigned int nthAttribute, int errorLevel)
 
 	snprintf(name, sizeof(name),
 			 PGrnLexiconNameFormat ".%s",
-			 index->rd_node.relNode,
+			 PGRN_RELATION_GET_LOCATOR_NUMBER(index),
 			 nthAttribute,
 			 PGrnIndexColumnName);
 	return PGrnLookup(name, errorLevel);
@@ -358,13 +360,13 @@ PGrnCreateTableWithSize(Relation index,
 	if (name)
 	{
 		flags |= GRN_OBJ_PERSISTENT;
-		if (index && index->rd_node.spcNode != MyDatabaseTableSpace)
+		if (index && PGRN_RELATION_GET_LOCATOR_SPACE(index) != MyDatabaseTableSpace)
 		{
 			char *databasePath;
 			char filePath[MAXPGPATH];
 
 			databasePath =
-				GetDatabasePath(MyDatabaseId, index->rd_node.spcNode);
+				GetDatabasePath(MyDatabaseId, PGRN_RELATION_GET_LOCATOR_SPACE(index));
 			snprintf(filePath, sizeof(filePath),
 					 "%s.%.*s",
 					 PGrnDatabaseBasename,
@@ -407,6 +409,100 @@ PGrnCreateTableWithSize(Relation index,
 	return table;
 }
 
+grn_obj *
+PGrnCreateSimilarTemporaryLexicon(Relation index, const char *tag)
+{
+	unsigned int i;
+	grn_obj *lexicon;
+	grn_table_flags flags = 0;
+	grn_obj *keyType = NULL;
+	grn_obj *tokenizer = NULL;
+	grn_obj *tokenizerBuffer = &(buffers->tokenizer);
+	grn_obj *normalizers = NULL;
+	grn_obj *normalizersBuffer = &(buffers->normalizers);
+	grn_obj *tokenFilters = NULL;
+	grn_obj *tokenFiltersBuffer = &(buffers->tokenFilters);
+	grn_obj *temporaryLexicon = NULL;
+
+	if (PGrnPGIsParentIndex(index))
+	{
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("pgroonga: %s PGrnCreateSimilarTemporaryLexicon: "
+						"index must not a parent index: <%s>",
+						tag,
+						RelationGetRelationName(index))));
+	}
+
+	for (i = 0; i < index->rd_att->natts; i++)
+	{
+		lexicon = PGrnLookupLexicon(index, i, ERROR);
+		if (grn_type_id_is_text_family(ctx, lexicon->header.domain))
+		{
+			break;
+		}
+		grn_obj_unref(ctx, lexicon);
+		lexicon = NULL;
+	}
+	if (!lexicon)
+	{
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("pgroonga: %s PGrnCreateSimilarTemporaryLexicon: "
+						"index doesn't have a lexicon for text: <%s>",
+						tag,
+						RelationGetRelationName(index))));
+	}
+
+	switch (lexicon->header.type)
+	{
+	case GRN_TABLE_HASH_KEY:
+		flags |= GRN_OBJ_TABLE_HASH_KEY;
+		break;
+	case GRN_TABLE_PAT_KEY:
+		flags |= GRN_OBJ_TABLE_PAT_KEY;
+		break;
+	case GRN_TABLE_DAT_KEY:
+		flags |= GRN_OBJ_TABLE_DAT_KEY;
+		break;
+	default:
+		break;
+	}
+
+	keyType = grn_ctx_at(ctx, lexicon->header.domain);
+
+	GRN_BULK_REWIND(tokenizerBuffer);
+	grn_table_get_default_tokenizer_string(ctx, lexicon, tokenizerBuffer);
+	if (GRN_TEXT_LEN(tokenizerBuffer) > 0) {
+		tokenizer = tokenizerBuffer;
+	}
+
+	GRN_BULK_REWIND(normalizersBuffer);
+	grn_table_get_normalizers_string(ctx, lexicon, normalizersBuffer);
+	if (GRN_TEXT_LEN(normalizersBuffer) > 0) {
+		normalizers = normalizersBuffer;
+	}
+
+	GRN_BULK_REWIND(tokenFiltersBuffer);
+	grn_table_get_token_filters_string(ctx, lexicon, tokenFiltersBuffer);
+	if (GRN_TEXT_LEN(tokenFiltersBuffer) > 0) {
+		tokenFilters = tokenFiltersBuffer;
+	}
+
+	temporaryLexicon = PGrnCreateTable(index,
+									   NULL,
+									   flags,
+									   keyType,
+									   tokenizer,
+									   normalizers,
+									   tokenFilters);
+
+	grn_obj_unref(ctx, lexicon);
+	grn_obj_unref(ctx, keyType);
+
+	return temporaryLexicon;
+}
+
 grn_obj *
 PGrnCreateColumn(Relation	index,
 				 grn_obj	*table,
@@ -437,7 +533,7 @@ PGrnCreateColumnWithSize(Relation	index,
 	if (name)
 	{
 		flags |= GRN_OBJ_PERSISTENT;
-		if (index && index->rd_node.spcNode != MyDatabaseTableSpace)
+		if (index && PGRN_RELATION_GET_LOCATOR_SPACE(index) != MyDatabaseTableSpace)
 		{
 			char *databasePath;
 			char tableName[GRN_TABLE_MAX_KEY_SIZE];
@@ -445,7 +541,7 @@ PGrnCreateColumnWithSize(Relation	index,
 			char filePath[MAXPGPATH];
 
 			databasePath =
-				GetDatabasePath(MyDatabaseId, index->rd_node.spcNode);
+				GetDatabasePath(MyDatabaseId, PGRN_RELATION_GET_LOCATOR_SPACE(index));
 			tableNameSize = grn_obj_name(ctx,
 										 table,
 										 tableName,
@@ -508,6 +604,21 @@ PGrnIndexColumnSetSourceIDs(Relation index,
 							grn_obj *indexColumn,
 							grn_obj *sourceIDs)
 {
+	grn_obj currentSourceIDs;
+	bool same;
+
+	GRN_RECORD_INIT(&currentSourceIDs, GRN_OBJ_VECTOR, GRN_ID_NIL);
+	grn_obj_get_info(ctx, indexColumn, GRN_INFO_SOURCE, &currentSourceIDs);
+	same = (GRN_BULK_VSIZE(&currentSourceIDs) == GRN_BULK_VSIZE(sourceIDs) &&
+			memcmp(GRN_BULK_HEAD(&currentSourceIDs),
+				   GRN_BULK_HEAD(sourceIDs),
+				   GRN_BULK_VSIZE(&currentSourceIDs)) == 0);
+	GRN_OBJ_FIN(ctx, &currentSourceIDs);
+	if (same)
+	{
+		return;
+	}
+
 	grn_obj_set_info(ctx, indexColumn, GRN_INFO_SOURCE, sourceIDs);
 	PGrnCheck("failed to set sources: <%s>: <%s>",
 			  PGrnInspectName(indexColumn),
@@ -554,6 +665,33 @@ PGrnRemoveObjectWithSize(const char *name,
 			  (int)nameSize, name);
 }
 
+void
+PGrnRemoveObjectForce(const char *name)
+{
+	PGrnRemoveObjectForceWithSize(name, strlen(name));
+}
+
+void
+PGrnRemoveObjectForceWithSize(const char *name,
+							  size_t nameSize)
+{
+	grn_obj *object;
+
+	object = PGrnLookupWithSize(name, nameSize, PGRN_ERROR_LEVEL_IGNORE);
+	if (object)
+	{
+		if (grn_obj_remove(ctx, object) != GRN_SUCCESS)
+		{
+			object = NULL;
+		}
+	}
+	if (!object) {
+		grn_obj_remove_force(ctx, name, nameSize);
+	}
+	PGrnCheck("failed to remove: <%.*s>",
+			  (int)nameSize, name);
+}
+
 void
 PGrnFlushObject(grn_obj *object, bool recursive)
 {
@@ -666,6 +804,9 @@ PGrnPGTypeToGrnType(Oid pgTypeID, unsigned char *flags)
 		typeID = GRN_DB_LONG_TEXT;
 		typeFlags |= GRN_OBJ_VECTOR;
 		break;
+	case UUIDOID:
+		typeID = GRN_DB_SHORT_TEXT;
+		break;
 	default:
 		PGrnCheckRC(GRN_FUNCTION_NOT_IMPLEMENTED,
 					"%s unsupported type: %u",
